/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: String.h,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */
#ifndef OpenSOAP_String_H
#define OpenSOAP_String_H

#include <OpenSOAP/ByteArray.h>

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

    typedef struct tagOpenSOAPString OpenSOAPString;
    typedef OpenSOAPString    *OpenSOAPStringPtr;


    int
    OPENSOAP_API
    OpenSOAPStringCreate(/* [out] */ OpenSOAPStringPtr *str);
    
    /* MultiByte str initialize */
    int
    OPENSOAP_API
    OpenSOAPStringCreateWithMB(/* [in]  */ const char *mb_str,
                               /* [out] */ OpenSOAPStringPtr *str);

    /* wchar_t str initialize */
    int
    OPENSOAP_API
    OpenSOAPStringCreateWithWC(/* [in]  */ const wchar_t *wc_str,
                               /* [out] */ OpenSOAPStringPtr *str);

    /* char encoding str initialize */
    int
    OPENSOAP_API
    OpenSOAPStringCreateWithCharEncodingString(/* [in]  */ const char *
											   char_enc,
                                               /* [in]  */ OpenSOAPByteArrayPtr
                                               char_enc_str,
                                               /* [out] */ OpenSOAPStringPtr *
											   str);

    /* UTF-8 encoding str initialize */
	int
	OPENSOAP_API
	OpenSOAPStringCreateWithUTF8(/* [in]  */ const char *utf8Str,
								 /* [out] */ OpenSOAPStringPtr *str);

    /* Use OpenSOAPString Resource retain.
     * this term from Objective-C */
    int
    OPENSOAP_API
    OpenSOAPStringRetain(/* [in] */ OpenSOAPStringPtr str);

    /* release OpenSOAPString Resource */
    int
    OPENSOAP_API
    OpenSOAPStringRelease(/* [in] */ OpenSOAPStringPtr str);

    /* get length for MultiByte */
    int
    OPENSOAP_API
    OpenSOAPStringGetLengthMB(/* [in]  */ OpenSOAPStringPtr str,
                              /* [out] */ size_t *len);
        
    /* get length for wchar_t */
    int
    OPENSOAP_API
    OpenSOAPStringGetLengthWC(/* [in]  */ OpenSOAPStringPtr str,
                              /* [out] */ size_t *len);

	/* get MultiByte string using memAllocator */
	int
	OPENSOAP_API
	OpenSOAPStringGetStringMBWithAllocator(/* [in]  */ OpenSOAPStringPtr
										   str,
										   /* [in]  */ char *
										   (*memAllocator)(size_t),
										   /* [out] */ size_t *len,
										   /* [out] */ char **mbStr);

	/* get wchar_t string using memAllocator */
	int
	OPENSOAP_API
	OpenSOAPStringGetStringWCWithAllocator(/* [in]  */ OpenSOAPStringPtr
										   str,
										   /* [in]  */ wchar_t *
										   (*memAllocator)(size_t),
										   /* [out] */ size_t *len,
										   /* [out] */ wchar_t **mbStr);

	int
	OPENSOAP_API
	OpenSOAPStringGetStringUTF8WithAllocator(/* [in]  */ OpenSOAPStringPtr
											 str,
											 /* [in]  */ char *
											 (*memAllocator)(size_t),
											 /* [out] */ size_t *len,
											 /* [out] */ char **utf8Str);

    /* get MultiByte string */
    int
    OPENSOAP_API
    OpenSOAPStringGetStringMB(/* [in]      */ OpenSOAPStringPtr str,
                              /* [in, out] */ size_t *len,
                              /* [out]     */ char * mb_str);
        
    /* get wchar_t string */
    int
    OPENSOAP_API
    OpenSOAPStringGetStringWC(/* [in]      */ OpenSOAPStringPtr str,
                              /* [in, out] */ size_t *len,
                              /* [out]     */ wchar_t *wc_str);

    /* get char encoding string */
    int
    OPENSOAP_API
    OpenSOAPStringGetCharEncodingString(/* [in]  */ OpenSOAPStringPtr str,
                                        /* [in]  */ const char *char_enc,
                                        /* [out] */ OpenSOAPByteArrayPtr 
                                        char_enc_str);
    
    /* set MultiByte string */
    int
    OPENSOAP_API
    OpenSOAPStringSetStringMB(/* [in] */ OpenSOAPStringPtr str,
                              /* [in] */ const char *mb_str);
        
    /* set wchar_t string */
    int
    OPENSOAP_API
    OpenSOAPStringSetStringWC(/* [in] */ OpenSOAPStringPtr str,
                              /* [in] */ const wchar_t *wc_str);

    /* set char encoding string */
    int
    OPENSOAP_API
    OpenSOAPStringSetCharEncodingString(/* [out] */ OpenSOAPStringPtr str,
                                        /* [in]  */ const char *char_enc,
                                        /* [in]  */ OpenSOAPByteArrayPtr 
                                        char_enc_str);

    /* set UTF-8 string */
    int
    OPENSOAP_API
    OpenSOAPStringSetStringUTF8(/* [out] */ OpenSOAPStringPtr str,
								/* [in]  */ const char *utf8Str);

    /* format string */
    int
    OPENSOAP_API
    OpenSOAPStringFormatMB(/* [out] */ OpenSOAPStringPtr str,
                           /* [in]  */ const char *format, ...);

    /* format string */
    int
    OPENSOAP_API
    OpenSOAPStringFormatWC(/* [out] */ OpenSOAPStringPtr str,
                           /* [in]  */ const wchar_t *format, ...);

    /* compare multibyte character string */
    int
    OPENSOAP_API
    OpenSOAPStringCompareMB(/* [in]  */ OpenSOAPStringPtr str,
                            /* [in]  */ const char *cmp_str,
                            /* [out] */ int *cmp_rslt);

    /* compare wide character string */
    int
    OPENSOAP_API
    OpenSOAPStringCompareWC(/* [in]  */ OpenSOAPStringPtr str,
                            /* [in]  */ const wchar_t *cmp_str,
                            /* [out] */ int *cmp_rslt);

    /* compare string */
    int
    OPENSOAP_API
    OpenSOAPStringCompare(/* [in]  */ OpenSOAPStringPtr str,
                          /* [in]  */ OpenSOAPStringPtr cmp_str,
                          /* [out] */ int *cmp_rslt);

    /* find string */
    int
    OPENSOAP_API
    OpenSOAPStringFindStringMB(/* [in]      */ OpenSOAPStringPtr str,
                               /* [in]      */ const char *find_str,
                               /* [in, out] */ size_t *idx);

    /* find string */
    int
    OPENSOAP_API
    OpenSOAPStringFindStringWC(/* [in]      */ OpenSOAPStringPtr str,
                               /* [in]      */ const wchar_t *find_str,
                               /* [in, out] */ size_t *idx);

    /* find string */
    int
    OPENSOAP_API
    OpenSOAPStringFindString(/* [in]      */ OpenSOAPStringPtr str,
                             /* [in]      */ OpenSOAPStringPtr find_str,
                             /* [in, out] */ size_t *idx);

    /* find string condition function */
    typedef int
    (*OpenSOAPStringFindIfFunc)(/* [in, out] */ void *opt,
                                /* [in]      */ wchar_t c,
                                /* [out]     */ int *judge);
	
    /* find string with condition function */
    extern
    int
    OPENSOAP_API
    OpenSOAPStringFindIfStringIndex(/* [in]  */ OpenSOAPStringPtr str,
                                    /* [in]  */ OpenSOAPStringFindIfFunc 
                                    find_func,
                                    /* [in]  */ void *find_func_opt,
                                    /* [in, out] */ size_t *idx);
    /* replace string */
    int
    OPENSOAP_API
    OpenSOAPStringReplaceStringMB(/* [in, out] */ OpenSOAPStringPtr str,
                                  /* [in]      */ const char *find_str,
                                  /* [in]      */ const char *rplc_str,
                                  /* [in, out] */ size_t *idx);
    
    /* replace string */
    int
    OPENSOAP_API
    OpenSOAPStringReplaceStringWC(/* [in, out] */ OpenSOAPStringPtr str,
                                  /* [in]      */ const wchar_t *find_str,
                                  /* [in]      */ const wchar_t *rplc_str,
                                  /* [in, out] */ size_t *idx);
    
    /* replace string */
    int
    OPENSOAP_API
    OpenSOAPStringReplaceString(/* [in, out] */ OpenSOAPStringPtr str,
                                /* [in]      */ OpenSOAPStringPtr find_str,
                                /* [in]      */ OpenSOAPStringPtr rplc_str,
                                /* [in, out] */ size_t *idx);
    
    /* clear OpenSOAPString */
    int
    OPENSOAP_API
    OpenSOAPStringClear(/* [out] */ OpenSOAPStringPtr str);

    /* concat MB string to OpenSOAPString */
    int
    OPENSOAP_API
    OpenSOAPStringAppendMB(/* [out] */ OpenSOAPStringPtr str,
                           /* [in]  */ const char *mb_str,
                           /* [in]  */ size_t       mb_len);

    /* concat MB string to OpenSOAPString */
    int
    OPENSOAP_API
    OpenSOAPStringAppendWC(/* [out] */ OpenSOAPStringPtr str,
                           /* [in]  */ const wchar_t *wc_str,
                           /* [in]  */ size_t   wc_len);

    /* duplicate string */
    int
    OPENSOAP_API
    OpenSOAPStringDuplicate(/* [in]  */ OpenSOAPStringPtr str,
                            /* [out] */ OpenSOAPStringPtr *dup_str);

    /* get substring */
    int
    OPENSOAP_API
    OpenSOAPStringGetSubstring(/* [in]  */ OpenSOAPStringPtr str,
                               /* [in]  */ size_t   beg,
                               /* [in]  */ size_t   len,
                               /* [out] */ OpenSOAPStringPtr *sub_str);
	
    /*
     * thise functions doesn't use OpenSOAPString member,
     * but this has something to do with String processing.
     */
    int
    OPENSOAP_API
    OpenSOAPStringConvertCharEncoding(/* [in]  */ const char *from_enc,
                                      /* [in]  */
									  OpenSOAPByteArrayPtr from_str,
                                      /* [in]  */ const char *to_enc,
                                      /* [out] */ OpenSOAPByteArrayPtr to_str);

	int
	OPENSOAP_API
	OpenSOAPStringConvertXMLCharRefToUTF8(/* [in]  */
										  const unsigned char *utf8Beg,
										  /* [out] */
										  OpenSOAPByteArrayPtr utf8BAry);
	int
	OPENSOAP_API
	OpenSOAPStringGetStringUSASCII(/* [in]  */ OpenSOAPStringPtr str,
								   /* [out] */ OpenSOAPByteArrayPtr charEncStr);

	int
	OPENSOAP_API
	OpenSOAPStringIterateProc(/* [in]  */ OpenSOAPStringPtr str,
							  /* [in]  */ int (*iterateProc)(unsigned long c,
															 size_t idx,
															 size_t len,
															 void *opt),
							  /* [in]  */ int (*beforeProc)(size_t len,
															void *opt),
							  /* [in]  */ int (*afterProc)(size_t len,
														   void *opt),
							  /* [in, out] */ void *opt);
	
#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* OpenSOAP_String_H */
