/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: ChannelManager.cpp,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#if HAVE_CONFIG_H
# include "config.h"
#endif

#if defined(WIN32)
#include <windows.h>
#include <process.h>
#else
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <pthread.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <string>
#include <iostream>

#include <stdexcept>

#include "ServerCommon.h"
#include "ChannelManager.h"

using namespace OpenSOAP;
using namespace std;

/*
  ChannelManager::
*/

//for LOG
static const std::string CLASS_SIG = "ChannelManager";

ChannelManager::ChannelManager()
{
#ifdef DEBUG
  cerr << CLASS_SIG << "::" << CLASS_SIG << endl;
#endif /* DEBUG */  
  setMaskFlag = false;
}

ChannelManager::~ChannelManager()
{
#ifdef DEBUG
  cerr << CLASS_SIG << "::~" << CLASS_SIG << endl;
#endif /* DEBUG */  

}

void
#if defined(WIN32)
ChannelManager::setMask(int newMask)
#else
ChannelManager::setMask(mode_t newMask)
#endif
{
  newMask_ = newMask;
  setMaskFlag = true;
}

bool 
ChannelManager::setSocketAddr(
#if defined(WIN32)
							const int addr)
#else
							const std::string& addr)
#endif
{
#ifdef DEBUG
  cerr << CLASS_SIG << "::setSocketAddr() arg=[" 
       << addr << "]" << endl;
#endif /* DEBUG */
  /*
  if (0 != unlink(addr.c_str())) {
    return false;
  }
  */
  socketAddr_ = addr;

  return true;
}

void ChannelManager::run()
{
  //===================================================================
  // ʐM ChannelManager <-> ChannelSelector
  //===================================================================
#if defined(WIN32)
	typedef SOCKADDR SockAddr;
	typedef SOCKADDR_IN SockAddrAf;
	//SOCKADDR_IN serverAddr;
	SockAddrAf serverAddr;
	SockAddrAf clientAddr;

	SOCKET sockfd = INVALID_SOCKET;
	SOCKET newsockfd = INVALID_SOCKET;
	WSADATA data;
	if (WSAStartup(0x101,&data)) {
		cerr << "** E **: WSAStartup failed." << endl;
		throw exception();
	}
	int af = AF_INET;
	int addrLen = sizeof(SockAddrAf);
	int clientLen = int();
#else

  typedef struct sockaddr SockAddr;
  typedef struct sockaddr_un SockAddrAf;

  int sockfd = int();
  int newsockfd = int();
  SockAddrAf serverAddr;
  SockAddrAf clientAddr;
  int af = AF_UNIX;
#if HAVE_TYPE_OF_ADDRLEN_AS_SOCKLEN_T
  socklen_t addrLen = socklen_t(); //socket_t();
  socklen_t clientLen = socklen_t();
#endif
#if HAVE_TYPE_OF_ADDRLEN_AS_SIZE_T
  // Solaris 2.6 does not have socklen_t
  size_t addrLen = size_t(); 
  size_t clientLen = size_t();
#endif
#if HAVE_TYPE_OF_ADDRLEN_AS_INT
  // Solaris 2.6 does not have socklen_t
  int addrLen = int(); 
  int clientLen = int();
#endif

#endif
  
  //ChannelSelectorƂ̒ʐM|[g
  if (0 > (sockfd = socket(af, SOCK_STREAM, 0))) {
    cerr << TAG_ERR << CLASS_SIG << "::" << CLASS_SIG
	 << ": open socket failed" << endl;
    //throw std::exception();
    // ONX̐݌vKv 
  }
#ifdef DEBUG
  cerr << CLASS_SIG << "::" << CLASS_SIG 
       << ": socket open OK!" << endl;
#endif

  memset((char*)&serverAddr, 0, sizeof(serverAddr));

#if defined(WIN32)
	serverAddr.sin_family = af;
	serverAddr.sin_addr.s_addr = INADDR_ANY;
	serverAddr.sin_port = htons(socketAddr_);
#else
  serverAddr.sun_family = af;
  strcpy(serverAddr.sun_path, socketAddr_.c_str());
  addrLen = sizeof(serverAddr.sun_family) + strlen(serverAddr.sun_path);

  unlink(socketAddr_.c_str());

  if (setMaskFlag == true) {
    curMask = umask(newMask_);
  }
#endif

  if (0 > bind(sockfd, (SockAddr*)&serverAddr, addrLen)) {
    cerr << TAG_ERR << CLASS_SIG << "::" << CLASS_SIG
	 << ": bind failed" << endl;
    //throw std::exception();
    // ONX̐݌vKv 
  }
#if !defined(WIN32)
  if (setMaskFlag == true) {
    umask(curMask);
  }
#endif
#ifdef DEBUG
  cerr << CLASS_SIG << "::" << CLASS_SIG 
       << ": bind OK!" << endl;
#endif

  if (0 > listen(sockfd, 5)) {
    cerr << TAG_ERR << CLASS_SIG << "::" << CLASS_SIG
	 << ": listen failed" << endl;
    //throw std::exception();
    // ONX̐݌vKv 
  }

#ifdef DEBUG
  cerr << CLASS_SIG << "::" << CLASS_SIG 
       << ": listen OK!" << endl;
#endif
    
  for (;;) {
    clientLen = sizeof(clientAddr);

#ifdef DEBUG
    cerr << CLASS_SIG << "::" << CLASS_SIG 
	 << ": accept ready!" << endl;
#endif
    
    newsockfd = accept(sockfd, (SockAddr*)&clientAddr, &clientLen);
    if (0 > newsockfd) {
      cerr << TAG_ERR << CLASS_SIG << "::" << CLASS_SIG
	   << ": accept failed" << endl;
      //throw std::exception();
      // ONX̐݌vKv 
    }

#ifdef DEBUG
    cerr << CLASS_SIG << "::" << CLASS_SIG 
	 << ": accept OK! fd=(" << newsockfd << ")" << endl;
#endif

    //XbhɓnKv̂f[^
    // Eint newsockfd;
    // EChannelManager* this;
    ThrData* thrData = new ThrData();
    thrData->that = this;
    thrData->sockfd = newsockfd;

    //ʐMXbh쐬
#if defined(WIN32)
	HANDLE thrId;
	thrId = (HANDLE)_beginthread(connectionThread, 0, thrData);
#else
    pthread_t thrId;
    int status = pthread_create(&thrId,
				NULL,
				connectionThread,
				(void*)thrData);
    if (0 != status) {
      cerr << TAG_ERR << CLASS_SIG << "::" << CLASS_SIG
	   << ": create thread failed" << endl;
      //throw std::exception();
      // ONX̐݌vKv 
    }
    pthread_detach(thrId);
#endif

  } // end for (;;)
}
#if defined(WIN32)
void
#else
void* 
#endif
ChannelManager::connectionThread(void* aArg)
{
#ifdef DEBUG
  cerr << CLASS_SIG << "::connectionThread()" << endl;
#endif /* DEBUG */

  //ñ`FbN
  if (!aArg) {
    cerr << TAG_ERR << CLASS_SIG 
	 << "::connectionThread: invalid argument" << endl;
#if defined(WIN32)
	_endthread();
#else
    pthread_exit(NULL);
#endif
  }

  //n̓WJ
  ThrData* thrData = (ThrData*)aArg;
  ChannelManager* that = thrData->that;
  int sockfd = thrData->sockfd;
  delete thrData;

#ifdef DEBUG
  cerr << CLASS_SIG << "::connectionThread: sockfd=(" << sockfd << ")" << endl;
#endif /* DEBUG */

  that->doProc(sockfd);

  if (0 < sockfd) {
#if defined(WIN32)
	closesocket(sockfd);
#else
    close(sockfd);
    sockfd = 0;
#endif
  }

#if defined(WIN32)
  _endthread();
#else
  pthread_exit(NULL);
  return NULL;
#endif

}

void 
ChannelManager::doProc(int sockfd)
{
#ifdef DEBUG
  cerr << CLASS_SIG << "::doProc() sockfd=[" << sockfd << "]" << endl;
#endif /* DEBUG */

}

