/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: Forwarder.cpp,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#if defined(_MSC_VER)
#include <process.h>
#else
#include <sys/time.h>
#include <unistd.h>
#endif

#include <iostream>
#include <algorithm>
#include <string>
#include <vector>

/* common library */
#include "connection.h"
#include "SOAPMessage.h"
#include "StringUtil.h"
#include "HTTPMessage.h"
#include "FileIDFunc.h"
#include "Timeout.h"
#include "FwdInfoFunc.h"
#include "SignatureFunc.h"

/* server common library */
#include "ServerCommon.h"
#include "ChannelDescriptor.h"

#include "TTLTableRefChnlSelector.h"

/* forwarder library */
#include "Forwarder.h"

using namespace OpenSOAP;
using namespace std;

#if defined(_MSC_VER)
int Forwarder::mtCount_ = 0;
#endif

Forwarder::Forwarder(const ChannelDescriptor& chnlDesc)
  : chnlDesc_(chnlDesc)
{
#ifdef DEBUG
  cerr << "Forwarder::Forwarder" << endl;
#endif

#if defined(_MSC_VER)
	if (mtCount_ > 10000) {
		mtCount_ = 0;
	}
	std::string mtId = "FORWARDERMTX" + StringUtil::toString(mtCount_++);
	_mt = CreateMutex(NULL, FALSE, mtId.c_str());
	//_mt = CreateMutex(NULL, FALSE, "FORWARDERMTX");
#else
  pthread_mutex_init(&_mt, NULL);
#endif

  defaultForwarding = getForwarderInfo(forwardHostname, forwardPort, 
				       forwardCGILocation);
  
  localhostName = getLocalhostName();
  std::vector<std::string> backwardURLs;
  bool hasBackwardURL = getBackwardURLs(backwardURLs);
  backwardURL = backwardURLs[0];
  if (!hasBackwardURL) {
    backwardURL = "http://localhost:80/soapInterface.cgi";
  }
}// end of Forwarder()

//----- Destructor -----//
Forwarder::~Forwarder()
{
#ifdef DEBUG
  cerr << "Forwarder::~Forwarder" << endl;
#endif

#if defined(_MSC_VER)
	CloseHandle(_mt);
#else
  pthread_mutex_destroy(&_mt);
#endif
}// end of ~Forwarder()

//----- bZ[W]Xbh -----//
#if defined(_MSC_VER)
void
#else
void*
#endif
Forwarder::syncMessageForward(void* arg)
{
  Forwarder* that = (Forwarder*)arg;

#if defined(_MSC_VER)
	WaitForSingleObject(that->_mt, INFINITE);
#else  
  int lock_status = pthread_mutex_lock(&(that->_mt));
  if (0 != lock_status) {
    cerr << "[[Error!!]] void* Forwarder::localSyncThrFunc(void*)" << endl;
    cerr << "            pthread_mutex_lock() failed." << endl;
    cerr.flush();
    return NULL;
  }
#endif

  // add signature if needed
  if(needSignature()) {
    if (!addSignatureToString(that->soapMsgFromMsgDrv)){
      cerr << "[[Error!!]] void* MsgDrv::localSyncThrFunc(void*)" << endl;
      cerr << "            addSignature failed." << endl;
    };
  }

  // HTTP bZ[W쐬B
  std::string http_msg 
    = createHTTPMessage(that->forwardCGILocation,
			that->soapMsgFromMsgDrv, that->operationName);

  // for debug  
#ifdef DEBUG_
  cout << "Forwarder::syncMessageForward() thr_func wait..." << endl;
#endif

  int fd = openSocket(that->forwardHostname, that->forwardPort);
  if (fd < 0) {
    cerr << "Error: socket [" << that->forwardHostname << ", "
	 << that->forwardPort << "] cannot open!" << endl;
  }
  sendSocket(fd, http_msg);
  //std::string recv_msg;
  recvSocket(fd, that->soapMsgFromOtherIf);

#ifdef DEBUG
cerr << "########## recvSocket !! ##########:[" 
	<< that->soapMsgFromOtherIf << "]" 
	<< endl;
#endif

#if defined(_MSC_VER)
	closesocket(fd);
#else
  close(fd);
#endif

  that->soapMsgFromOtherIf = removeHTTPHeader(that->soapMsgFromOtherIf);

  // mutex ̉
#if defined(_MSC_VER)
	ReleaseMutex(that->_mt);
#else
  int unlock_status = pthread_mutex_unlock(&(that->_mt));
  if (0 != unlock_status) {
    cerr << "[[Error!!]] void* Forwarder::localSyncThrFunc(void*)" << endl;
    cerr << "            pthread_mutex_unlock() failed." << endl;
    cerr.flush();
    return NULL;
  }
#endif

#ifdef DEBUG
cerr << "############# END!! ###########" << endl;
#endif

#if defined(_MSC_VER)
	_endthread();
#else
	pthread_exit(NULL);
  return NULL;
#endif

}

//----- bZ[W] -----//
bool
Forwarder::execSyncMessageForward()
{
#ifdef DEBUG_
  cout << "[execSyncMessageForward]" << endl;
  cout << "forwardHostname = " << forwardHostname << endl;
  cout << "forwardPort = " << forwardPort << endl;
  cout << "forwardCGILocation = " << forwardCGILocation << endl;
  cout << "operationName = " << operationName << endl;
#endif

#if !defined(_MSC_VER)
  struct timeval tv_start, tv_end;
  struct timezone tz;
  
  gettimeofday(&tv_start, &tz);
  cerr << "tv_start.tv_sec  = " << tv_start.tv_sec << endl;
  cerr.flush();
#endif
  
  bool ttlHeaderExist;
  std::string ttlType;
  getTTLInfo(soapMsgFromMsgDrv, ttlHeaderExist, ttl, ttlType);
#ifdef DEBUG
  // for debug
  cerr << "Forwarder::execSyncMessageForward() ttl=(" << ttl << ")" << endl;
#endif
  // for debug
#ifdef DEBUG_
  cout << "Forwarder::execSyncMessageForward() create thread..." << endl;
#endif

  //ړȈsȂXbh𐶐
#if defined(_MSC_VER)
	HANDLE thread_t;
	thread_t = (HANDLE)_beginthread(syncMessageForward, 0, this);
#else
  pthread_t thread_t;
  int status = pthread_create(&thread_t, NULL, syncMessageForward, this);

#ifdef DEBUG_
  // for debug
  cout << "Forwarder::execSyncMessageForward() " 
       << "thread create done... status=(" << status << ")" << endl;
#endif

#endif

  //^C}[IuWFNg
//for test
	ttl = 30;
  OpenSOAP::Timeout timeout(ttl, thread_t);
  //^CAEgĎJn
  timeout.start();
  
  //ړȈɃXbh
#if defined(_MSC_VER)
	if (WaitForSingleObject(thread_t, INFINITE) == STATUS_WAIT_0) {
#ifdef DEBUG
	    cerr << "Forwarder::execSyncMessageForward() thread joined" << endl;
#endif
	}
	else {
	    cerr << "Forwarder::execSyncMessageForward() thread join error" 
			<< endl;
	}
#else
  status = pthread_join(thread_t, NULL);
  
  if (status != 0) {
    cerr << "Forwarder::execSyncMessageForward() pthread join error" << endl;
  }
#endif

  //ړIIɂ^C}[~
  timeout.finish();

#ifdef DEBUG_  
  // for debug
  cout << "Forwarder::execSyncMessageForward() bye..." << endl;
  cout << "timeout.targetThreadIsCanceled = " 
       << timeout.targetThreadIsCanceled() << endl;
#endif

#if !defined(_MSC_VER)
  gettimeofday(&tv_end, &tz);
  cerr << "tv_end.tv_sec  = " << tv_end.tv_sec << endl;
  cerr.flush();
#endif
  
  return timeout.targetThreadIsCanceled();
}

bool
Forwarder::setFwdHostInfo()
{
#ifdef DEBUG_
  cout << "----- setFwdHostInfo() -----" << endl;
  cout << "localhostName = " << localhostName << endl;
#endif

  unsigned int i = 0;
  for (i = 0; i < hostnames.size() - 1; i++) {
    if (localhostName == hostnames[i]) {
      forwardHostname = hostnames[i + 1];
      forwardPort = ports[i + 1];
      forwardCGILocation = cgiLocations[i + 1];
      break;
    }
  }
  
  if (localhostName == *(hostnames.end() - 1)) {
    return false;
  }
  
  if (i == hostnames.size() - 1) {
    forwardHostname = hostnames[0];
    forwardPort = ports[0];
    forwardCGILocation = cgiLocations[0];
  }
  
#ifdef DEBUG_  
  cout << "forwardHostname = " << forwardHostname << endl;
  cout << "forwardPort = " << forwardPort << endl;
  cout << "forwardCGILocation = " << forwardCGILocation << endl;
  cout << "----- setFwdHostInfo() -----" << endl;
#endif

  return true;
}// end of setFwdHostInfo()

//----- 2 ڈȍ~̃zXgł, received_path ǉ ------//
bool
Forwarder::addReceivedPath () {
  for (unsigned i = 0; i < hostnames.size(); i++) {
    if (hostnames[i] == localhostName) {
      attachReceivedPath(soapMsgFromMsgDrv, 
			 fileIDOfMsgFromMsgDrv, backwardURL);
      return true;
    }
  }// loop of for (i)
  
  return false;
}// end of attatchReceivePath()

//----- TTL Table 烁bZ[W ID ɑΉ -----/
//----- backward_path T -----/
bool
Forwarder::refToTTLTable(const std::string& msgIDContent, 
			 std::string& backwardPath)
{
  ChannelSelector *ttlTableRefChnlSelector 
    = new TTLTableRefChnlSelector();
  ChannelDescriptor ttlTableRefChnlDesc;
  
  if (0 != ttlTableRefChnlSelector->open(ttlTableRefChnlDesc)) {
    cerr << "Forwarder::refToTTLTable: ttlTableRefChnl open failed" << endl;
    return false;
  }

#ifdef DEBUG_
  cout << "send to ttl table" << endl;
#endif

  if (0 > ttlTableRefChnlDesc.write(msgIDContent)) {
    cerr << "Forwarder::refToTTLTable: write failed to TTLTableRef" << endl;
    return false;
  }
  
#ifdef DEBUG_
  cout << "recv from ttl table" << endl;  
#endif

  if (0 > ttlTableRefChnlDesc.read(backwardPath)) {
    cerr << "Forwarder::refToTTLTable: read failed from TTLTableRef" << endl;
    return false;
  }
  cerr << "** MsgDrv ** READ From TTLTable=[" 
       << backwardPath << "]" << endl;
  
  ttlTableRefChnlSelector->close(ttlTableRefChnlDesc);
  delete ttlTableRefChnlSelector;
  
  
  return true;
}// end of refToTTLTable()

//----- X|X߂s -----//
bool
Forwarder::execBackwarding()
{
  cerr << "----- Forwarder::execBackwarding begin -----" << endl;
  std::string backwardPath;
  std::string msgIDContent = getMsgIDContent(soapMsgFromMsgDrv);
  bool hasBackwardPath = refToTTLTable(msgIDContent, backwardPath);
  
  if (hasBackwardPath) {
    // TTL e[u backward pXꍇ֖͂߂D
    std::string prtcl;
    analyseFwdHostInfo(backwardPath, prtcl, forwardHostname,
		       forwardPort, forwardCGILocation);
    
    bool targetThreadIsCanceled = execSyncMessageForward();    
    if (targetThreadIsCanceled) {
      // ^CAEg thr_func LZďIꍇB
      soapMsgFromOtherIf = makeTimeoutMessage();
    }
    
  }
  cerr << "----- Forwarder::execBackwarding end   -----" << endl;
  return true;
}// end of execBackwarding()

void
Forwarder::run()
{

  cerr << "FORWARDER::run read ready." << endl;
  
  if (0 > chnlDesc_.read(fileIDOfMsgFromMsgDrv)) {
    cerr << TAG_ERR << "Forwarder::run: read failed" << endl;
    return;
  }

  cerr << "$$ Forwarder $$ READ From MsgDrv=[" 
       << fileIDOfMsgFromMsgDrv << "]" << endl;

  soapMsgFromMsgDrv = revertFromFileID(fileIDOfMsgFromMsgDrv);
  
  // forwarder  operationNames, ttl nĂ炦΁A
  // 
  cerr << "extract Operation and Parameter" << endl;
  std::vector<std::string> operationNames;
  std::vector<std::string> paramNames;
  std::vector<std::string> paramValues;

  bool operationExist 
    = extractOperationAndParam(soapMsgFromMsgDrv, operationNames, 
			       paramNames, paramValues);
  // ܂ł̏͂ȂB

  // SOAP Header ]ɊւW߂
  bool fwdPathExist;
  std::vector<std::string> fwdPath;
  bool hopCountExist;
  unsigned int hopCount;
  getFwdInfoFromHeader(soapMsgFromMsgDrv, fwdPathExist, fwdPath,
		       hopCountExist, hopCount);
  
  // X|XbZ[Wǂ𒲂ׂ
  bool itIsResMsg = isItResponseMessage(soapMsgFromMsgDrv);
  
  bool targetThreadIsCanceled = false;

  if (itIsResMsg) {
    execBackwarding();
  }

  //2002/10/08 update
  if ((!defaultForwarding && !fwdPathExist) ||
	(hopCountExist && hopCount <= 0)) {
	// reply FaultMessagge 
    soapMsgFromOtherIf = makeOperationNotFoundMessage();
  }
  else {
	//forward Next Host
    ttl = 5;
    if (fwdPathExist) {
      analyseFwdHostInfo(fwdPath, prtcls, hostnames, ports, cgiLocations);
      addReceivedPath();
      updateBackwardPath(soapMsgFromMsgDrv, backwardURL);

	  setFwdHostInfo();
	}
    if (hopCountExist) {
		decreaseHopCount(soapMsgFromMsgDrv);
	}
    targetThreadIsCanceled = execSyncMessageForward();

	if (targetThreadIsCanceled) {
		soapMsgFromOtherIf = makeTimeoutMessage();
	}
  }

  fileIDOfMsgFromOtherIf = convertToFileID(soapMsgFromOtherIf);
  
  cerr << "$$ Forwarder $$ WRITE To MsgDrv=[" 
       << fileIDOfMsgFromOtherIf << "]" << endl;
  
  if (0 > chnlDesc_.write(fileIDOfMsgFromOtherIf)) {
    cerr << TAG_ERR << "Forwarder::run: write failed" << endl;
    return;
  }
}

