/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: MsgDrv.cpp,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#if defined(WIN32)
#include <process.h>
#else
#include <sys/time.h>
#include <unistd.h>
#endif

#include <iostream>

#include "ServerCommon.h"
#include "ChannelDescriptor.h"
#include "SrvDrvChnlSelector.h"
#include "ForwarderChnlSelector.h"
#include "FwdQueuePushChnlSelector.h"
#include "ReqQueuePushChnlSelector.h"
#include "ResSpoolPopChnlSelector.h"
#include "ResSpoolPushChnlSelector.h"
#include "TTLTablePushChnlSelector.h"
#include "TTLTableRefChnlSelector.h"
#include "MsgDrv.h"
#include "SignatureFunc.h"
#include "StringUtil.h"

/* common library */
#include "connection.h"
#include "FileIDFunc.h"
#include "SOAPMessage.h"
#include "Timeout.h"

/* SSML Information Manager library */
#include "SSMLInfoFunc.h"
#include "SSMLInfo.h"

/* Forwarder Information Function library */
#include "FwdInfoFunc.h"

using namespace OpenSOAP;
using namespace std;


#if defined(WIN32)
int MsgDrv::mtCount_ = 0;
#endif

//----- Constructor -----//
MsgDrv::MsgDrv(const ChannelDescriptor& chnlDesc)
  : chnlDesc_(chnlDesc)
{
#ifdef DEBUG
  cerr << "MsgDrv::MsgDrv" << endl;
#endif

#if defined(WIN32)
	if (mtCount_ > 10000) {
		mtCount_ = 0;
	}
	std::string mtId = "MSGDRVMTX" + StringUtil::toString(mtCount_++);
	_mt = CreateMutex(NULL, FALSE, mtId.c_str());
	//_mt = CreateMutex(NULL, FALSE, "MSGDRVMTX");
#else
  pthread_mutex_init(&_mt, NULL);
#endif

  downsideTimeout = 5;
  
  std::string  forwardHostname;
  unsigned int forwardPort; 
  std::string  forwardCGILocation;
  
  defaultForwarding 
    = getForwarderInfo(forwardHostname, forwardPort, 
		       forwardCGILocation);
  
  bool localURLExist = getBackwardURLs(localURLs);
  if (!localURLExist) {
    cerr << "local URL not Exist" << endl;
  }
  
}// end of MsgDrv (const ChannelDescriptor&)

//----- Destructor -----//
MsgDrv::~MsgDrv ()
{
#ifdef DEBUG
  cerr << "MsgDrv::~MsgDrv" << endl;
#endif

#if defined(WIN32)
	CloseHandle(_mt);
#else
  pthread_mutex_destroy(&_mt);
#endif
}// end of ~MsgDrv ()

//----- SrvDrv ւ̑Ms߂̃Xbh  -----//
#if defined(WIN32)
void
#else
void*
#endif 
MsgDrv::localSyncThrFunc (void* arg)
{
  MsgDrv* that = (MsgDrv*)arg;

#ifdef DEBUG
  // for debug  
  cout << "MsgDrv::localSyncThrFunc() thr_func wait..." << endl;
#endif

  // Service Driver ̕ԓ҂
  
  ChannelDescriptor srvDrvChnlDesc;
  ChannelSelector* srvDrvChnlSelector = new SrvDrvChnlSelector(false);

  if (0 != srvDrvChnlSelector->open(srvDrvChnlDesc)) {

    cerr << "MsgDrv::run: SrvDrvChnl open failed" << endl;
#if defined(WIN32)
	_endthread();
#else
	pthread_exit(NULL);
	return NULL;
#endif
  }



#if defined(WIN32)
	WaitForSingleObject(that->_mt, INFINITE);
#else  
  int lock_status = pthread_mutex_lock (&(that->_mt));
  if (0 != lock_status) {
    cerr << "[[Error!!]] void* MsgDrv::localSyncThrFunc(void*)" << endl;
    cerr << "            pthread_mutex_lock() failed." << endl;
    cerr.flush();
	pthread_exit(NULL);
    return NULL;
  }
#endif



  if (0 > srvDrvChnlDesc.write(that->fileIDOfMsgFromIf)) {

    cerr << "MsgDrv::run: write failed to SrvDrv" << endl;
#if defined(WIN32)
	_endthread();
#else
	pthread_exit(NULL);
	return NULL;
#endif
  }



  if (0 > srvDrvChnlDesc.read(that->fileIDOfMsgFromSrvdrv)) {

    cerr << "MsgDrv::run: read failed from SrvDrv" << endl;
#if defined(WIN32)
	_endthread();
#else
	pthread_exit(NULL);
	return NULL;
#endif

  }

//#ifdef DEBUG
  cerr << "** MsgDrv ** READ From SrvDrv=[" 
       << that->fileIDOfMsgFromSrvdrv << "]" << endl;
//#endif

#if defined(WIN32)
	ReleaseMutex(that->_mt);
#else
  int unlock_status = pthread_mutex_unlock (&(that->_mt));
  if (0 != unlock_status) {
    cerr << "[[Error!!]] void* MsgDrv::localSyncThrFunc(void*)" << endl;
    cerr << "            pthread_mutex_unlock() failed." << endl;
    cerr.flush();
	pthread_exit(NULL);
    return NULL;
  }
#endif


  srvDrvChnlSelector->close(srvDrvChnlDesc);

  delete srvDrvChnlSelector;


  if(needSignature()) {
    std::string msgFromSrvDrv 
      = revertFromFileID(that->fileIDOfMsgFromSrvdrv);
    if (!addSignatureToString(msgFromSrvDrv)){
      cerr << "[[Error!!]] void* MsgDrv::localSyncThrFunc(void*)" << endl;
      cerr << "            addSignatureToString failed." << endl;
    }
    else {
      that->fileIDOfMsgFromSrvdrv = convertToFileID(msgFromSrvDrv);
    }
  }
  
  // write Response Messege of SOAP Interface
  that->chnlDesc_.write(that->fileIDOfMsgFromSrvdrv); 

#if defined(WIN32)
	_endthread();
#else
  pthread_exit(NULL);
  return 0;
#endif
}

//----- tH[hs߂̃Xbh  -----//
#if defined(WIN32)
void
#else
void*
#endif
MsgDrv::fwdThrFunc (void* arg)
{
  MsgDrv* that = (MsgDrv*)arg;
  
#ifdef DEBUG
  // for debug  
  cout << "MsgDrv::fwdThrFunc() wait..." << endl;
#endif

  // Forwarder ̕ԓ҂
  
  ChannelDescriptor fwderChnlDesc;
  ChannelSelector* fwderChnlSelector = new ForwarderChnlSelector();
  if (0 != fwderChnlSelector->open(fwderChnlDesc)) {
    cerr << "MsgDrv::run: FwderChnl open failed" << endl;
#if defined(WIN32)
	_endthread();
#else
	pthread_exit(NULL);
    return NULL;
#endif
  }

#ifdef DEBUG
cerr << "******* fwder open done. ***********" << endl;
#endif


#if defined(WIN32)
	WaitForSingleObject(that->_mt, INFINITE);
#else  
  int lock_status = pthread_mutex_lock (&(that->_mt));
  if (0 != lock_status) {
    cerr << "[[Error!!]] void* MsgDrv::fwdThrFunc(void*)" << endl;
    cerr << "            pthread_mutex_lock() failed." << endl;
    cerr.flush();
	pthread_exit(NULL);
    return NULL;
  }
#endif

#ifdef DEBUG
cerr << "******* fwder write... ***********" << endl;
#endif

  if (0 > fwderChnlDesc.write(that->fileIDOfMsgFromIf)) {
    cerr << "MsgDrv::run: write failed to Fwder" << endl;
#if defined(WIN32)
	_endthread();
#else
	pthread_exit(NULL);
    return NULL;
#endif
  }

#ifdef DEBUG
cerr << "******* fwder write done. ***********" << endl;
#endif

  if (0 > fwderChnlDesc.read(that->fileIDOfMsgFromSrvdrv)) {
    cerr << "MsgDrv::run: read failed from Fwder" << endl;
#if defined(WIN32)
	_endthread();
#else
	pthread_exit(NULL);
    return NULL;
#endif
  }

//#ifdef DEBUG
  cerr << "** MsgDrv ** READ From Fwder=[" 
       << that->fileIDOfMsgFromSrvdrv << "]" << endl;
//#endif

#if defined(WIN32)
	ReleaseMutex(that->_mt);
#else
  int unlock_status = pthread_mutex_unlock (&(that->_mt));
  if (0 != unlock_status) {
    cerr << "[[Error!!]] void* MsgDrv::fwdThrFunc(void*)" << endl;
    cerr << "            pthread_mutex_unlock() failed." << endl;
    cerr.flush();
	pthread_exit(NULL);
    return NULL;
  }
#endif

  fwderChnlSelector->close(fwderChnlDesc);
  delete fwderChnlSelector;
  
  that->chnlDesc_.write(that->fileIDOfMsgFromSrvdrv); 
  
#if defined(WIN32)
	_endthread();
#else
  pthread_exit(NULL);
  return NULL;
#endif
}

//----- SrvDrv ւ̑M܂̓tH[hs -----//
void* 
MsgDrv::sendToSrvDrvOrFwd()
{
#if !defined(WIN32)  
  struct timeval tv_start, tv_end;
  struct timezone tz;
  
  gettimeofday(&tv_start, &tz);

  cerr << "tv_start.tv_sec  = " << tv_start.tv_sec << endl;
  cerr.flush();
#endif



#ifdef DEBUG
  // for debug
  cerr << "MsgDrv::sendToSrvDrvOrFwd() ttl =(" << ttl << ")" << endl;
#endif


#ifdef DEBUG
  // for debug
  cout << "MsgDrv::sendToSrvDrvOrFwd() create thread..." << endl;
#endif

  //ړȈsȂXbh𐶐
#if defined(WIN32)
	HANDLE
#else
  pthread_t 
#endif
	thread_t;

  int status = 0;
  if (operationExist) {

#if defined(WIN32)
	thread_t = (HANDLE)_beginthread(localSyncThrFunc, 0, this);
#else
    status = pthread_create(&thread_t, NULL, localSyncThrFunc, this);
#endif

  }
  else {

    ttl = 5;
    //ttl = 35;
#if defined(WIN32)
	thread_t = (HANDLE)_beginthread(fwdThrFunc, 0, this);
#else
    status = pthread_create(&thread_t, NULL, fwdThrFunc, this);
#endif

  }
  
#ifdef DEBUG
  // this debug
  cout << "MsgDrv::sendToSrvDrvOrFwd() " 
       << "thread create done..." << endl;
#endif

  //^C}[IuWFNg
  OpenSOAP::Timeout timeout(ttl, thread_t);

  //^CAEgĎJn
  timeout.start();


  //ړȈɃXbh
#if defined(WIN32)
	if (WaitForSingleObject(thread_t, INFINITE) == STATUS_WAIT_0) {
#ifdef DEBUG
	    cerr << "MsgDrv::sendToSrvDrvOrFwd() thread joined" << endl;
#endif

	}
	else {

	    cerr << "MsgDrv::sendToSrvDrvOrFwd() thread join error" << endl;
	}
#else
  status = pthread_join(thread_t, NULL);
  if (status != 0) {
    cerr << "MsgDrv::sendToSrvDrvOrFwd() pthread join error" << endl;
  }
#endif

  //ړIIɂ^C}[~
  timeout.finish();

#ifdef DEBUG
  // for debug
  cout << "MsgDrv::sendToSrvDrvOrFwd() bye..." << endl;
  cout << "timeout.targetThreadIsCanceled = " 
       << timeout.targetThreadIsCanceled() << endl;
#endif

#if !defined(WIN32)
  gettimeofday(&tv_end, &tz);
  cerr << "tv_end.tv_sec  = " << tv_end.tv_sec << endl;
  cerr.flush();
#endif

  return (void*)timeout.targetThreadIsCanceled();

}


void
MsgDrv::popFromResponseSpool(std::string fileIDOfMsgID)
{
  ChannelDescriptor resSpoolPopChnlDesc;
  ChannelSelector* resSpoolPopChnlSelector = new ResSpoolPopChnlSelector();
  if (0 != resSpoolPopChnlSelector->open(resSpoolPopChnlDesc)) {
    cerr << "MsgDrv::popFromResponseSpool() fifo open failed" << endl;
    std::string errorMsg = makeInvalidRequestMessage();
    std::string fileIDOfErrorMsg = convertToFileID(errorMsg);
    chnlDesc_.write(fileIDOfErrorMsg);
  }
  
#ifdef DEBUG
  cout << "send msgID to ress" << endl;
#endif

  if (0 > resSpoolPopChnlDesc.write(fileIDOfMsgID)) {
    cerr << "MsgDrv::popFromResponseSpool: "
	 << "write failed to ResSpoolPop" << endl;
    return;
  }

//#ifdef DEBUG
  cerr << "** MsgDrv ** WRITE To ResSpoolPop=[" 
       << fileIDOfMsgID << "]" << endl;
//#endif

  std::string fileIDOfMsgFromResS;
#ifdef DEBUG
  cout << "recv from ttl table" << endl;  
#endif

  if (0 > resSpoolPopChnlDesc.read(fileIDOfMsgFromResS)) {
    cerr << "[[Error!!]] MsgDrv::popFromResponseSpool: "
	 << "            read failed from ResSpoolPop" << endl;
    return;
  }

//#ifdef DEBUG
  cerr << "** MsgDrv ** READ From ResSpoolPop=[" 
       << fileIDOfMsgFromSrvdrv << "]" << endl;
//#endif

  resSpoolPopChnlSelector->close(resSpoolPopChnlDesc);
  delete resSpoolPopChnlSelector;

  // add signatur if needed
  if(needSignature()) {
    std::string msgFromResS = revertFromFileID(fileIDOfMsgFromResS);
    if (!addSignatureToString(msgFromResS)){
      cerr << "[[Error!!]] MsgDrv::popFromResponseSpool: " << endl;
      cerr << "            addSignatureToString failed." << endl;
    }
    else {
      fileIDOfMsgFromResS = convertToFileID(msgFromResS);
    }
  }
  
  chnlDesc_.write(fileIDOfMsgFromResS);   

  std::string msgFromResS;
  msgFromResS = revertFromFileID(fileIDOfMsgFromResS);
#ifdef DEBUG
  cout << "msg from ress = " << msgFromResS << endl;
#endif

}

//----- TTL Table ɋL^ -----/
bool
MsgDrv::pushToTTLTable()
{
  ChannelSelector *ttlTablePushChnlSelector 
    = new TTLTablePushChnlSelector();
  ChannelDescriptor ttlTablePushChnlDesc;
  
  if (0 != ttlTablePushChnlSelector->open(ttlTablePushChnlDesc)) {
    cerr << "MsgDrv::pushToTTLTable: ttlTablePushChnl open failed" << endl;
    return false;
  }
  
#ifdef DEBUG
  cout << "send to ttl table" << endl;
#endif

  if (0 > ttlTablePushChnlDesc.write(fileIDOfMsgFromIf)) {
    cerr << "MsgDrv::pushToTTLTable: write failed to TTLTablePush" << endl;
    return false;
  }
  
  std::string fileIDOfMsgFromTTLTable;
#ifdef DEBUG
  cout << "recv from ttl table" << endl;  
#endif

  if (0 > ttlTablePushChnlDesc.read(fileIDOfMsgFromTTLTable)) {
    cerr << "MsgDrv::pushToTTLTable: read failed from TTLTablePush" << endl;
    return false;
  }

//#ifdef DEBUG
  cerr << "** MsgDrv ** READ From TTLTable=[" 
       << fileIDOfMsgFromTTLTable << "]" << endl;
//#endif

  ttlTablePushChnlSelector->close(ttlTablePushChnlDesc);
  delete ttlTablePushChnlSelector;
  
  return true;
}// end of pushToTTLTable()

bool 
MsgDrv::pushToQueue(bool operationExist, bool fwdPathExist, 
		    bool fwdHopExist, bool defaultForwarding)
{
  // bZ[W ID ̕t
  std::string msgID = attachMsgID(soapMsgFromIf, fileIDOfMsgFromIf, 
				  operationNames[0]);
  
  // bZ[W ID tꂽV SOAP bZ[W̃t@C ID 擾
  fileIDOfMsgFromIf = convertToFileID(soapMsgFromIf);

  // add signatur to MsgID, if needed
  if(needSignature()) {
    if (!addSignatureToString(msgID)) {
      cerr << "[[Error!!]] void* MsgDrv::pushToQueue()" << endl;
      cerr << "            addSignatureToString failed." << endl;
    };
  }
  // convert to File ID
  std::string fileIDOfMsgID = convertToFileID(msgID);

  // write to SOAP Interface
  chnlDesc_.write(fileIDOfMsgID);

  pushToTTLTable();
  
  ChannelDescriptor queuePushChnlDesc;
  ChannelSelector* queuePushChnlSelector = NULL;

  if (operationExist) {
    // Iy[Vꍇ
    if (fwdPathExist) {
      // ]oHw肳Ăꍇ
      bool itIsAsyncFinalReceiver 
	= isItAsyncFwdFinalReceiver(soapMsgFromIf, localURLs);
      if (itIsAsyncFinalReceiver) {
	// ŏI̎󂯎̏ꍇ
	queuePushChnlSelector = new ReqQueuePushChnlSelector();
      }
      else {
	// ]oH̓ȑꍇ
	queuePushChnlSelector = new FwdQueuePushChnlSelector();
      }
    }
    else {
      // Iy[VA]oHw肳ĂȂꍇ
      queuePushChnlSelector = new ReqQueuePushChnlSelector();
    }
  }
  else {
    // Iy[VȂꍇ
    if (fwdPathExist) {
      // ]oHꍇ
      bool itIsAsyncFinalReceiver 
	= isItAsyncFwdFinalReceiver(soapMsgFromIf, localURLs);
      if (itIsAsyncFinalReceiver) {
	// ŏI̎󂯎̏ꍇ
	queuePushChnlSelector = new ReqQueuePushChnlSelector();
      }
      else {
	// ŏI̎󂯎łȂꍇ
	queuePushChnlSelector = new FwdQueuePushChnlSelector();
      }
    } 
    else if (fwdHopExist) {
      // 䐔w]ꍇ
      queuePushChnlSelector = new FwdQueuePushChnlSelector();
    } 
    else {
      // Iy[VȂA]Ȃꍇ
      OpenSOAP::SSMLInfo ssmlInfo;
      soapMsgFromSrvdrv
	= createSOAPFaultMessage("Server.NoOperationAndNoForward",
				 "no operation found and no forward found",
				 "server", ssmlInfo.getOperationName() 
				 + " not respond");
      return true;
    }
  }
  
  if (queuePushChnlSelector == NULL) {
      OpenSOAP::SSMLInfo ssmlInfo;
      soapMsgFromSrvdrv
	= createSOAPFaultMessage("Server.NoOperationAndNoForward",
				 "no operation found and no forward found",
				 "server", ssmlInfo.getOperationName() 
				 + " not respond");
      return true;
  }
  
  if (0 != queuePushChnlSelector->open(queuePushChnlDesc)) {
    cerr << "MsgDrv::pushToReqQueue: queuePushChnl open failed" << endl;
    return false;
  }
  
#ifdef DEBUG
  cout << "send to reqq" << endl;
#endif

  if (0 > queuePushChnlDesc.write(fileIDOfMsgFromIf)) {
    cerr << "MsgDrv::pushToReqQueue: write failed to ReqQueuePush" << endl;
    return false;
  }
  
  std::string fileIDOfMsgFromQueue;
#ifdef DEBUG
  cout << "recv from reqq" << endl;  
#endif

  if (0 > queuePushChnlDesc.read(fileIDOfMsgFromQueue)) {
    cerr << "MsgDrv::pushToReqQueue: read failed from ReqQueuePush" << endl;
    return false;
  }

//#ifdef DEBUG
  cerr << "** MsgDrv ** READ From ReqQueuePush=[" 
       << fileIDOfMsgFromQueue << "]" << endl;
//#endif

  queuePushChnlSelector->close(queuePushChnlDesc);
  delete queuePushChnlSelector;

//======================================  
//  chnlDesc_.write(fileIDOfMsgFromSrvdrv); 
//======================================  
  
  std::string msgFromQueue;
  msgFromQueue = revertFromFileID(fileIDOfMsgFromQueue);
#ifdef DEBUG
  cout << "msg from reqq = " << msgFromQueue << endl;
#endif

  return false;
}

//----- TTL Table 烁bZ[W ID ɑΉ -----/
//----- backward_path T -----/
/* ̃\bh Forwarder ɂ̂łƂ瓝 */
bool
MsgDrv::refToTTLTable(const std::string& msgIDContent, 
		      std::string& backwardPath)
{
  ChannelSelector *ttlTableRefChnlSelector 
    = new TTLTableRefChnlSelector();
  ChannelDescriptor ttlTableRefChnlDesc;
  
  if (0 != ttlTableRefChnlSelector->open(ttlTableRefChnlDesc)) {
    cerr << "MsgDrv::refToTTLTable: ttlTableRefChnl open failed" << endl;
    return false;
  }
  
#ifdef DEBUG
  cout << "send to ttl table" << endl;
#endif

  if (0 > ttlTableRefChnlDesc.write(msgIDContent)) {
    cerr << "MsgDrv::refToTTLTable: write failed to TTLTableRef" << endl;
    return false;
  }
  
  
#ifdef DEBUG
  cout << "recv from ttl table" << endl;  
#endif

  if (0 > ttlTableRefChnlDesc.read(backwardPath)) {
    cerr << "MsgDrv::refToTTLTable: read failed from TTLTableRef" << endl;
    return false;
  }

//#ifdef DEBUG
  cerr << "** MsgDrv ** READ From TTLTable=[" 
       << backwardPath << "]" << endl;
//#endif

  ttlTableRefChnlSelector->close(ttlTableRefChnlDesc);
  delete ttlTableRefChnlSelector;
  
  
  return true;
}// end of refToTTLTable()

//----- X|XX|XbZ[Wɂ -----//
bool
MsgDrv::pushToResponseSpool(std::string fileIDOfMsgFromTransIF)
{
  ChannelDescriptor resSpoolPushChnlDesc;
  ChannelSelector* resSpoolPushChnlSelector;
  
  resSpoolPushChnlSelector = new ResSpoolPushChnlSelector();
  
  if (0 != resSpoolPushChnlSelector->open(resSpoolPushChnlDesc)) {
    cerr << "MsgDrv::PushToResponseSpool: Chnl open failed" << endl;
    return false;
  }
  
#ifdef DEBUG
  cout << "send to response spool" << endl;
#endif

  if (0 > resSpoolPushChnlDesc.write(fileIDOfMsgFromTransIF)) {
    cerr << "MsgDrv::PushToResponseSpool: write failed to Service" << endl;
    return false;
  }
  
  std::string fileIDOfMsgFromResSMgr;
  if (0 > resSpoolPushChnlDesc.read(fileIDOfMsgFromResSMgr)) {
    cerr << "MsgDrv::PushToResponseSpool: read failed from Service" << endl;
    return false;
  }
  
//#ifdef DEBUG
  cerr << "** MsgDrv ** READ From ResSpool=[" 
       << fileIDOfMsgFromResSMgr << "]" << endl;
//#endif

  resSpoolPushChnlSelector->close(resSpoolPushChnlDesc);
  delete resSpoolPushChnlSelector;
  
  
#ifdef DEBUG
  cout << "send to trans if" << endl;
#endif

  std::string msgToReqQMgr("success");
  std::string fileIDOfMsgToReqQMgr = convertToFileID(msgToReqQMgr);
  if (0 > chnlDesc_.write(fileIDOfMsgToReqQMgr)) {
    cerr << TAG_ERR << "MsgDrv::pushToResponseSpool: write failed" << endl;
    return false;
  }
  
  return true;
  
}// end of pushToResponseSpool(std::string fileIDOfMsgFromSrvdrv)

bool
MsgDrv::run()
{
#ifdef DEBUG
  cerr << "MsgDrv::run()" << endl;
#endif

  // TransIF 瑗Ăf[^ (t@C ID) ǂݍށB
  if (0 > chnlDesc_.read(fileIDOfMsgFromIf)) {

    cerr << TAG_ERR << "run: read failed" << endl;
    return false;
  }

//#ifdef DEBUG
  cerr << "** MsgDrv ** READ From TransIF=[" 
       << fileIDOfMsgFromIf << "]" << endl;
//#endif

  // t@C ID t@C̓eɕϊB
  soapMsgFromIf = revertFromFileID (fileIDOfMsgFromIf);

  // SOAP bZ[W̎ނ̃`FbN
  bool soapMsgIsValid;

 
  // bZ[W ID H
  bool itIsMsgID = false;
  soapMsgIsValid = isItMsgID(soapMsgFromIf, itIsMsgID);
#ifdef DEBUG
  cerr << "itIsMsgID = " << itIsMsgID << endl;
#endif

  if (!soapMsgIsValid) {

    soapMsgFromSrvdrv = makeInvalidRequestMessage();
    fileIDOfMsgFromSrvdrv = convertToFileID(soapMsgFromSrvdrv);
    chnlDesc_.write(fileIDOfMsgFromSrvdrv); 
    return false;
  }

  if (itIsMsgID) {

    std::string fileIDOfMsgID = fileIDOfMsgFromIf;
    popFromResponseSpool(fileIDOfMsgID);
    return true;
  }

  // <response>  true ŁA
  // <backward_path>  TTLMgr ɓo^ĂȂƂ
  // X|XXv[ɑD
  bool itIsResMsg = isItResponseMessage(soapMsgFromIf);

  std::string msgIDContent = getMsgIDContent(soapMsgFromIf);

  std::string backwardPath;
  if (msgIDContent.size() > 0) {

    refToTTLTable(msgIDContent, backwardPath);
    if (itIsResMsg && backwardPath == "") {
#ifdef DEBUG
      cout << "pushToResponseSpool()" << endl;
#endif
      pushToResponseSpool(fileIDOfMsgFromIf);
      return true;
    }
  }

  // Iy[V͉H
#ifdef DEBUG
  cerr << "extract Operation and Parameter" << endl;
#endif

  string nsStr;
  soapMsgIsValid 
    = extractOperationAndParam (soapMsgFromIf, operationNames, 
				paramNames, paramValues, &nsStr);
  //				paramNames, paramValues);

#ifdef DEBUG
  cerr << "##### MsgDrv:(" << soapMsgIsValid << "): "
       << "nsStr=[" << nsStr << "]" << endl;
#endif

  if (!soapMsgIsValid) {

    std::string invalidReqErrMsg = makeInvalidRequestMessage();
    std::string fileIDOfInvalidReqErrMsg = convertToFileID(invalidReqErrMsg);
    chnlDesc_.write(fileIDOfInvalidReqErrMsg); 
    return false;
  }

  /* Does async header exist ? */
  // 񓯊wb_tĂ邩H
  bool asyncHeaderExist, isAsync;

  soapMsgIsValid 
    = getAsyncInfoFromHeader(soapMsgFromIf, asyncHeaderExist, isAsync);

  if (!soapMsgIsValid) {

    std::string invalidReqErrMsg = makeInvalidRequestMessage();
    std::string fileIDOfInvalidReqErrMsg = convertToFileID(invalidReqErrMsg);
    chnlDesc_.write(fileIDOfInvalidReqErrMsg); 
    return false;
  }

  /* Does fwd header exist ? */
  // ]wb_tĂ邩H
  bool fwdPathExist, fwdHopExist;
  std::vector<std::string> fwdPath;
  unsigned int fwdHop;
  
  soapMsgIsValid 
    = getFwdInfoFromHeader(soapMsgFromIf, fwdPathExist, fwdPath,
			   fwdHopExist, fwdHop);

  if (!soapMsgIsValid) {
    std::string invalidReqErrMsg = makeInvalidRequestMessage();
    std::string fileIDOfInvalidReqErrMsg = convertToFileID(invalidReqErrMsg);
    chnlDesc_.write(fileIDOfInvalidReqErrMsg); 
    return false;
  }
  
#ifdef DEBUG
  cerr << "fwdPathExist = " << fwdPathExist << endl;
  cerr << "fwdHopExist = " << fwdHopExist << endl;
#endif

  /* Does ttl header exist ? */
  bool ttlHeaderExist;
  ttl = 0;
  std::string ttlType;

  soapMsgIsValid 
    = getTTLInfo(soapMsgFromIf, ttlHeaderExist, ttl, ttlType);

  if (!soapMsgIsValid) {
    std::string invalidReqErrMsg = makeInvalidRequestMessage();
    std::string fileIDOfInvalidReqErrMsg = convertToFileID(invalidReqErrMsg);
    chnlDesc_.write(fileIDOfInvalidReqErrMsg); 
    return false;
  }

//    bool isFwdByPathName;
//    std::vector<std::string> fwdPathName;
//    bool isFwdByHopCount;
//    unsigned int fwdHopCountValue;
  
  //* SSML p[^𒲂ׂ */
  OpenSOAP::SSMLInfo ssmlInfo;
  ssmlInfo.setOperationName(operationNames[0]);
  ssmlInfo.setNamespace(nsStr);
  operationExist = getSSMLInfo(ssmlInfo);

#ifdef DEBUG
  cerr << "### MsgDrv: operationExist=(" << operationExist << ")" << endl;
  cerr << "MsgDrv::run::"
       << "ssmlInfo.getSyncTTL() = " << ssmlInfo.getSyncTTL() << endl;
#endif

  bool targetThreadIsCanceled = false;
  
  if (!operationExist && !defaultForwarding && !fwdPathExist) {

    // Iy[V݂A
    // ftHg̃tH[hȂ
    // tH[hwb_Ȃꍇ
    std::string oprNotFoundMsg = makeOperationNotFoundMessage();
    std::string fileIDOfOprNotFoundMsg = convertToFileID(oprNotFoundMsg);
    chnlDesc_.write(fileIDOfOprNotFoundMsg); 
    return false;
  }
  
  if (!asyncHeaderExist || !isAsync) {

    /* async wb_ȂD܂ async wb_ false ̏ꍇ */
    /* (̏ꍇ) */
    if (operationExist || defaultForwarding || fwdPathExist) {
      // Iy[V݂D܂̓ftHg̃tH[h悪ꍇ
      if (!ttlHeaderExist) {
		ttl = ssmlInfo.getSyncTTL();
      }  

      targetThreadIsCanceled = sendToSrvDrvOrFwd();

    }
    if (targetThreadIsCanceled) {

      // ^CAEg thr_func LZďIꍇB
      soapMsgFromSrvdrv = makeTimeoutMessage();
      fileIDOfMsgFromSrvdrv = convertToFileID(soapMsgFromSrvdrv);
      chnlDesc_.write(fileIDOfMsgFromSrvdrv); 
    }
  }
  else {

    /* async wb_ true ̏ꍇ */
    /* (񓯊̏ꍇ) */
    if (operationExist || defaultForwarding || fwdPathExist || fwdHopExist) {

      // Iy[V݂D܂̓ftHg̃tH[h悪ꍇ
      //        targetThreadIsCanceled = sendToSrvDrvOrFwd();
      if (!ttlHeaderExist) {
		ttl = ssmlInfo.getAsyncTTL();
      }

      targetThreadIsCanceled 
		= pushToQueue(operationExist, fwdPathExist, 
			      fwdHopExist, defaultForwarding);
    }
    if (targetThreadIsCanceled) {

      // ^CAEg thr_func LZďIꍇB
      soapMsgFromSrvdrv = makeTimeoutMessage();
      fileIDOfMsgFromSrvdrv = convertToFileID(soapMsgFromSrvdrv);
      chnlDesc_.write(fileIDOfMsgFromSrvdrv); 
    }
    
  }

  return true;
  
}// end of run()
