/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: SrvDrvCreator_main.cpp,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#if defined(WIN32)
#include <windows.h>
#include <io.h>
#else
#include <unistd.h>
#endif

#include <iostream>
#include <fstream>
#include <string>

#if !defined(NO_LOGGER)
#include "Logger.h"
#endif
#include "ServerCommon.h"

#include "SrvDrvCreator.h"

using namespace OpenSOAP;
using namespace std;

int 
main(int argc, char** argv) {
  bool isAsync = false;
  
  // It works as Asynchronous Service Driver, 
  // except for argument is 0 or no argument exists

	// initial argument differ in windows and other.
#if defined(WIN32)
	// windows argument : program logfile [isAsyncFlg]
	string logFile = OPENSOAP_TMP_PATH + "log/srvDrvCreator_sync";
	if (argc > 2) {
	    if (atoi(argv[2]) != 0) {
			isAsync = true;
			logFile = OPENSOAP_TMP_PATH + "log/srvDrvCreator_async";
	    }
	}
#else
	// other argument : program [isAsyncFlg]
  if (argc > 1) {
    if (atoi(argv[1]) != 0) {
      isAsync = true;
    }
  }
#endif

#if defined(NO_LOGGER)
	if (argc >1) {
		logFile = argv[1];
	}
	FILE* outfp = fopen(logFile.c_str(), "w");
	if (!outfp) {
		return 1;
	}
	if (argc <=1) {
		fprintf(outfp, "invalid initial argument. argc=(%d)\n", argc);
		fprintf(outfp, "exit.\n");
		fclose(outfp);
		return EXIT_FAILURE;
	}
#if 1
	int old;
	int olderr;
	//backup stdout
	old = _dup(1);
	//stdout -> outfp
	if (-1 == _dup2(_fileno(outfp), 1)) {
		fprintf(outfp, "_dup2 to stdout failed.\n");
		fprintf(outfp, "exit.\n");
		fclose(outfp);
		return EXIT_FAILURE;
	}
	//backup stderr
	olderr = _dup(2);
	//stderr -> outfp
	if (-1 == _dup2(_fileno(outfp), 2)) {
		fprintf(outfp, "_dup2 to stderr failed.\n");
		fprintf(outfp, "exit.\n");
		fclose(outfp);
		return EXIT_FAILURE;
	}
#endif
#else
  // set log level  
  Logger::InitOpt opt;
  opt.outputLevel = LOG_DEBUG;
  Logger::init(opt);
  //write start message
  if (isAsync) {
    Logger::printLog(NULL, " SrvDrvCreator_main (async) start", LOG_INFO,
		     0x00100003L, "exe");
  }
  else {
    Logger::printLog(NULL, " SrvDrvCreator_main (sync) start", LOG_INFO,
		     0x00100003L, "exe");
  }
#endif

#if !defined(WIN32)
  // make it to daemon & write pid file
#if HAVE_DAEMON
  daemon(0, 1);
#endif
  
  pid_t pid = getpid();
  
  ofstream pidFile;
  if (isAsync) {
    // Asynchronous Communication
    cout << "Async SrvDrvCreator" << endl;
    pidFile.open(ASYNC_SRVDRV_PID_FILE.c_str());
    if (!pidFile) {
      cerr << "Error: pid file [" 
	   << ASYNC_SRVDRV_PID_FILE << "] cannot open!" << endl;
      exit(1);
    }
    pidFile << pid;
    pidFile.close();
  }
  else {
    // Synchronous Communication
    cout << "Sync SrvDrvCreator" << endl;
    pidFile.open(SYNC_SRVDRV_PID_FILE.c_str());
    if (!pidFile) {
      cerr << "Error: pid file [" 
	   << SYNC_SRVDRV_PID_FILE << "] cannot open!" << endl;
      exit(1);
    }
  }
  
  pidFile << pid;
  pidFile.close();
#endif

  // create instance of QueueManager and run
  ChannelManager* chnlMngr = new SrvDrvCreator(isAsync);
  chnlMngr->run();

  delete chnlMngr;

#if defined(NO_LOGGER)
#if 1
	fclose(outfp);
	_dup2(old, 1);
	_dup2(olderr, 2);
#endif
#else
  //write stop message
  if (isAsync) {
    Logger::printLog(NULL, " SrvDrvCreator_main (async) stop", LOG_INFO,
		     0x00100003L, "exe");
  }
  else {
    Logger::printLog(NULL, " SrvDrvCreator_main (sync) stop", LOG_INFO,
		     0x00100003L, "exe");
  }
  Logger::term();
#endif

  return EXIT_SUCCESS;
}
